#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <c10/cuda/CUDAGuard.h>

// Define warp size and tile size: each warp will process TILE_SIZE output features simultaneously.
#define WARP_SIZE 32
#define TILE_SIZE 4

// Kernel to compute GEMM with bias and ReLU activation using optimized global memory loads.
// Global memory loads from x, weight, and bias use __ldg() for read-only accesses and 128-bit aligned loads through float4.

__global__ void warp_tile_ldg_opt_kernel(const float* __restrict__ x,
                                          const float* __restrict__ weight,
                                          const float* __restrict__ bias,
                                          float* __restrict__ out,
                                          int in_features,
                                          int out_features) {
    // Each block corresponds to one batch row
    int batch_idx = blockIdx.x;

    // Determine warp-related indices
    int warps_per_block = blockDim.x / WARP_SIZE;  // Number of warps per block
    int warp_id_in_block = threadIdx.x / WARP_SIZE;  
    int lane = threadIdx.x % WARP_SIZE;

    // Each warp processes TILE_SIZE output features concurrently.
    int base_out = (blockIdx.y * warps_per_block + warp_id_in_block) * TILE_SIZE;

    // Early exit if the base output feature index is out of range
    if (base_out >= out_features) return;

    // Array to hold partial sums for each of the TILE_SIZE output features
    float sums[TILE_SIZE] = {0.0f, 0.0f, 0.0f, 0.0f};

    // Pointer to the current input row for the batch
    const float* x_row = x + batch_idx * in_features;
    // Use vectorized pointer for 128-bit aligned loads
    const float4* x_vec = reinterpret_cast<const float4*>(x_row);
    int vec_count = in_features / 4;          // Number of full float4 elements
    int remainder = in_features - vec_count * 4; // Number of remaining floats

    // Process each output feature in the current tile
    #pragma unroll
    for (int t = 0; t < TILE_SIZE; t++) {
        int out_idx = base_out + t;
        if (out_idx >= out_features) break;
        
        // Pointer to the corresponding weight row, use __restrict__ and assume 128-bit alignment
        const float* w_row = weight + out_idx * in_features;
        const float4* w_vec = reinterpret_cast<const float4*>(w_row);
        
        float local_sum = 0.0f;
        
        // Process main vectorized part. Each thread processes elements in a strided loop over the float4 blocks.
        for (int i = lane; i < vec_count; i += WARP_SIZE) {
            float4 x_val = __ldg(x_vec + i);
            float4 w_val = __ldg(w_vec + i);
            local_sum += x_val.x * w_val.x + x_val.y * w_val.y + x_val.z * w_val.z + x_val.w * w_val.w;
        }

        // Process any remaining scalar elements
        int rem_start = vec_count * 4;
        for (int i = rem_start + lane; i < in_features; i += WARP_SIZE) {
            local_sum += __ldg(x_row + i) * __ldg(w_row + i);
        }

        sums[t] = local_sum;
    }

    // Warp-level reduction using shuffle operations for each tile output
    #pragma unroll
    for (int t = 0; t < TILE_SIZE; t++) {
        for (int offset = WARP_SIZE / 2; offset > 0; offset /= 2) {
            sums[t] += __shfl_down_sync(0xffffffff, sums[t], offset);
        }
    }

    // The first thread in the warp writes the final results
    if (lane == 0) {
        #pragma unroll
        for (int t = 0; t < TILE_SIZE; t++) {
            int out_idx = base_out + t;
            if (out_idx < out_features) {
                float result = sums[t] + __ldg(bias + out_idx);
                // Apply ReLU activation
                out[batch_idx * out_features + out_idx] = (result > 0.0f) ? result : 0.0f;
            }
        }
    }
}


// Host function to launch the kernel
torch::Tensor warp_tile_ldg_opt_forward(torch::Tensor x,
                                          torch::Tensor weight,
                                          torch::Tensor bias) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");

    int batch_size = x.size(0);
    int in_features = x.size(1);
    int out_features = weight.size(0);

    auto out = torch::empty({batch_size, out_features}, x.options());

    // Configure grid and block dimensions.
    // Each block handles one batch row. The grid's y-dimension covers all output features in chunks of (warps_per_block * TILE_SIZE).
    int warps_per_block = 8;  // Tunable parameter
    int threads_per_block = warps_per_block * WARP_SIZE;
    int grid_y = (out_features + (warps_per_block * TILE_SIZE) - 1) / (warps_per_block * TILE_SIZE);

    dim3 grid(batch_size, grid_y);
    dim3 block(threads_per_block);

    cudaStream_t stream = c10::cuda::getCurrentCUDAStream();
    warp_tile_ldg_opt_kernel<<<grid, block, 0, stream>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        in_features,
        out_features
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &warp_tile_ldg_opt_forward, "GEMM with bias and ReLU using __ldg and 128-bit aligned loads (CUDA)");
}
